<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DDC3699;

use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\ManyToOne;
use Doctrine\ORM\Mapping\Table;

#[Table(name: 'ddc3699_relation_many')]
#[Entity]
class DDC3699RelationMany
{
    /** @var int */
    #[Id]
    #[Column(type: 'integer')]
    public $id;

    /** @var DDC3699Child */
    #[ManyToOne(targetEntity: 'DDC3699Child', inversedBy: 'relations')]
    public $child;
}
