<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DDC3231;

use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\GeneratedValue;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\Table;

#[Table(name: 'users')]
#[Entity(repositoryClass: 'DDC3231User1Repository')]
class DDC3231User1
{
    /** @var int */
    #[Id]
    #[Column(type: 'integer')]
    #[GeneratedValue(strategy: 'AUTO')]
    protected $id;

    /** @var string */
    #[Column(type: 'string', length: 255)]
    protected $name;
}
