<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\ValueConversionType;

use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\JoinColumn;
use Doctrine\ORM\Mapping\ManyToOne;
use Doctrine\ORM\Mapping\Table;

#[Table(name: 'vct_owning_manytoone_extralazy')]
#[Entity]
class OwningManyToOneExtraLazyEntity
{
    /** @var string */
    #[Column(type: 'rot13', length: 255)]
    #[Id]
    public $id2;

    /** @var InversedOneToManyExtraLazyEntity */
    #[ManyToOne(targetEntity: 'InversedOneToManyExtraLazyEntity', inversedBy: 'associatedEntities')]
    #[JoinColumn(name: 'associated_id', referencedColumnName: 'id1')]
    public $associatedEntity;
}
