<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DDC3293;

class DDC3293UserPrefixed
{
    /** @var string */
    protected $id;

    /** @var DDC3293Address */
    protected $address;
}
