<?php

declare(strict_types=1);

namespace Doctrine\ORM\Cache;

/**
 * Cache entry interface
 *
 * <b>IMPORTANT NOTE:</b>
 *
 * Fields of classes that implement CacheEntry are public for performance reason.
 */
interface CacheEntry
{
}
