<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DDC3293;

class DDC3293Address
{
    /** @var string */
    public $street;

    /** @var string */
    public $city;

    /** @var string */
    public $country;
}
