<?php

declare(strict_types=1);

namespace Doctrine\ORM\Cache\Exception;

use Doctrine\ORM\Cache\CacheException as BaseCacheException;

/**
 * Exception for cache.
 */
class CacheException extends BaseCacheException
{
}
