<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DDC3579;

use Doctrine\ORM\Mapping\AssociationOverride;
use Doctrine\ORM\Mapping\AssociationOverrides;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Mapping\Entity;

#[Entity]
#[AssociationOverrides([new AssociationOverride(name: 'groups', inversedBy: 'admins')])]
class DDC3579Admin extends DDC3579User
{
    public static function loadMetadata(ClassMetadata $metadata): void
    {
        $metadata->setAssociationOverride('groups', ['inversedBy' => 'admins']);
    }
}
